package ch.epfl.tpp;

import java.io.IOException;
import java.io.InputStream;

public class Base64InputStream extends InputStream {
    private final static String decodingString = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

    private final InputStream encodedStream;
    
    /* Invariant: buffer contains, in its least-significant bits, bufferLen
     * decoded bits that still have to be sent to the client. Since bits are
     * decoded in chunks of 6 bits and sent to the client in chunks of 8,
     * bufferLen will rotate between the values 0, 4, and 2.
     */
    private int buffer = 0;
    private int bufferLen = 0;

    public Base64InputStream(InputStream underlyingStream) {
        this.encodedStream = underlyingStream;
    }
    
    @Override
    public int read() throws IOException {
        assert bufferLen == 0 || bufferLen == 2 || bufferLen == 4;
        // Read and decode enough bits from the encoded stream to be able to
        // send 8 back.
        while (bufferLen < 8) {
            int encodedChar = encodedStream.read();
            if (encodedChar == -1) {
                // If non-zero decoded bits are in the buffer when the encoded
                // stream dries up, there is an decoding error.
                if (bufferLen > 0 && (buffer & ((1 << bufferLen) - 1)) != 0)
                    throw new IOException(bufferLen + " non-zero decoded bits left at EOS!");
                return -1;
            }
            assert 0x00 <= encodedChar && encodedChar <= 0xFF;
            // Ignore padding characters.
            if (encodedChar != '=') {
                int decodedBits = decodingString.indexOf(encodedChar);
                if (decodedBits == -1)
                    throw new IOException("invalid base64 character: "+ (char)encodedChar);
                buffer = (buffer << 6) | decodedBits;
                bufferLen += 6;
            }
        }
        // We now have at least 8 bits in the buffer, so we send the 8 most
        // significant ones back to the client and keep the rest in buffer.
        bufferLen -= 8;
        return (buffer >> bufferLen) & 0xFF;
    }
    
    @Override
    public void close() throws IOException {
        encodedStream.close();
    }
}
